"""Additionally restrict the maximum drive speed near drive range limits.

If you understand the junior-high physics, you know the algorithm.

"""

import astropy.units as u

from ..typing import QuantityValue, Unit
from .. import utils


class Decelerate:
    def __init__(
        self,
        limit: utils.ValueRange[u.Quantity],
        max_acceleration: u.Quantity,
    ) -> None:
        self.limit = limit
        self.max_acceleration = max_acceleration

    def __call__(
        self,
        encoder_reading: QuantityValue,
        velocity: QuantityValue,
        angle_unit: Unit = None,
    ) -> u.Quantity:
        velocity = utils.get_quantity(
            velocity, unit=(None if angle_unit is None else f"{angle_unit}/s")
        )
        encoder_reading = utils.get_quantity(encoder_reading, unit=angle_unit)

        if encoder_reading not in self.limit:
            return 0 << velocity.unit

        position_relative_to_limits = (limit - encoder_reading for limit in self.limit)
        max_velocity_squared = (
            2 * self.max_acceleration * rel for rel in position_relative_to_limits
        )
        max_velocity = (
            v**0.5 if v.value == 0 else v / (abs(v) ** 0.5)
            for v in max_velocity_squared
        )

        return utils.clip(velocity, *max_velocity)



__all__ = ["configure", "config"]

import difflib
import os
import shutil
from collections import UserDict
from pathlib import Path
from types import SimpleNamespace
from typing import Any, Callable, Dict, List

import astropy.units as u
from astropy.coordinates import EarthLocation
from tomlkit.toml_file import TOMLFile

from . import EnvVarName, get_logger
from .utils import ValueRange

logger = get_logger(__name__)


class _ConfigParsers(UserDict):
    def __missing__(self, key: str) -> Callable[[Any], Any]:
        if key in self.keys():
            return self[key]
        similar_keys = difflib.get_close_matches(key.lower(), self.keys(), cutoff=0.85)
        if similar_keys:
            logger.info(
                f"Parser for '{key}' not found, instead using raw value. "
                f"Similarly named parameters : {similar_keys}"
            )
        return lambda x: x


class Configuration:
    """NECST configuration.

    This parameter collection supports flexible name look-up. If you query the parameter
    'antenna' by ``config.antenna``, all parameters prefixed by 'antenna' will be
    extracted.

    Examples
    --------
    >>> neclib.config.observatory
    'OMU1P85M'
    >>> neclib.config.antenna_pid_param_az
    [1.5, 0.0, 0.0]
    >>> neclib.config.antenna_pid_param
    SimpleNamespace(az=[1.5, 0.0, 0.0], el=[1.5, 0.0, 0.0])
    >>> neclib.config.antenna_pid_param.az
    [1.5, 0.0, 0.0]

    """

    _instance = None

    DefaultNECSTRoot = Path.home() / ".necst"
    DefaultConfigPath = DefaultNECSTRoot / "config.toml"

    def __new__(cls):
        if cls._instance is None:
            return super().__new__(cls)
        return cls._instance

    def __init__(self):
        self.__parameters = []
        self.reload()

    def __repr__(self) -> str:
        length = max(len(p) for p in self.__parameters)

        def _prettify(key: str):
            value = self.__dict__.get(key, None)
            return f"    {key:{length+2}s}{value!s}    ({type(value).__name__})"

        _parameters = "\n".join([_prettify(k) for k in self.__parameters])
        return f"NECST configuration\n{_parameters}"

    def __str__(self) -> str:
        def _format(key: str):
            value = self.__dict__[key]
            return f"{key}={value!s}"

        _parameters = ", ".join([_format(k) for k in self.__parameters])
        return f"Configuration({_parameters})"

    def reload(self):
        """Reload the parameters, in case the config file is updated.

        .. WARNING::

            Manually set parameters will be lost on reload.

        """
        for param in self.__parameters:
            del self.__dict__[param]
        self.__config_path = self.__find_config_path()
        self.__dotnecst = self.__config_path.parent
        self.__parameters = self.__parse()

    def __getattr__(self, key: str) -> Any:
        prefix = key + "_"
        prefix_length = len(prefix)
        match = {
            k[prefix_length:]: getattr(self, k)
            for k in self.__parameters
            if k.startswith(prefix)
        }
        return SimpleNamespace(**match) if match else None

    def __get_parser(self) -> _ConfigParsers:
        _parsers: Dict[str, Callable[[Any], Any]] = {
            "observatory": str,
            "location": lambda x: EarthLocation(**x),
            "simulator": bool,
            "antenna_pid_param_az": list,
            "antenna_pid_param_el": list,
            "antenna_drive_range_az": lambda x: ValueRange(*map(u.Quantity, x)),
            "antenna_drive_range_el": lambda x: ValueRange(*map(u.Quantity, x)),
            "antenna_drive_warning_limit_az": lambda x: ValueRange(*map(u.Quantity, x)),
            "antenna_drive_warning_limit_el": lambda x: ValueRange(*map(u.Quantity, x)),
            "antenna_drive_critical_limit_az": lambda x: ValueRange(
                *map(u.Quantity, x)
            ),
            "antenna_drive_critical_limit_el": lambda x: ValueRange(
                *map(u.Quantity, x)
            ),
            "antenna_pointing_accuracy": u.Quantity,
            "antenna_pointing_parameter_path": lambda x: self.__dotnecst / Path(x),
            "antenna_max_acceleration_az": u.Quantity,
            "antenna_max_acceleration_el": u.Quantity,
            "antenna_max_speed_az": u.Quantity,
            "antenna_max_speed_el": u.Quantity,
            "antenna_speed_to_pulse_factor_az": u.Quantity,
            "antenna_speed_to_pulse_factor_el": u.Quantity,
            "antenna_command_frequency": int,
            "antenna_command_offset_sec": float,
            "ros_service_timeout_sec": float,
            "ros_communication_deadline_sec": float,
        }
        return _ConfigParsers(_parsers)

    def __parse(self) -> List[str]:
        raw_config = TOMLFile(self.__config_path).read().unwrap()
        parser = self.__get_parser()
        for k, v in raw_config.items():
            setattr(self, k, parser[k](v))
        return list(raw_config.keys())

    def __find_config_path(self) -> Path:
        candidates = [self.DefaultNECSTRoot]
        if EnvVarName.necst_root in os.environ.keys():
            candidates.insert(0, Path(os.environ[EnvVarName.necst_root]))

        for path in candidates:
            config_path = path if path.is_file() else path / "config.toml"
            if config_path.exists():
                logger.info(f"Imported configuration file '{config_path}'")
                return config_path
        logger.error(
            "Config file not found, using the default parameters. "
            "To create the file with default parameters, run `neclib.configure()`."
        )
        return Path(__file__).parent / "src" / "config.toml"

    @classmethod
    def configure(cls) -> None:
        """Create config file under ``$HOME/.necst``"""
        cls.DefaultNECSTRoot.mkdir(exist_ok=True)
        for filename in ["config.toml", "pointing_param.toml"]:
            _target_path = cls.DefaultNECSTRoot / filename
            if _target_path.exists():
                logger.error(f"'{_target_path}' already exists, skipping...")
                continue
            shutil.copyfile(Path(__file__).parent / "src" / filename, _target_path)
        cls().reload()


configure = Configuration.configure
config = Configuration()



"""Informative exception, regarding the NECST system status."""

__all__ = ["NECSTAuthorityError", "ConfigurationError"]


class NECSTAuthorityError(Exception):
    """Error related to controlling priority.

    Examples
    --------
    >>> raise neclib.NECSTAuthorityError("reason or advise to fix this error")

    """

    pass


class ConfigurationError(Exception):
    """Error related to parameter configuration."""

    pass



from types import ModuleType
from typing import Any, Dict, List

from neclib import config, get_logger
from ..exceptions import ConfigurationError

logger = get_logger(__name__)


def parse_device_configuration(modules: List[ModuleType]) -> Dict[str, Any]:
    implementations = list_implementations(modules)

    devices = config.dev
    if devices is None:
        raise ConfigurationError("No device configuration found.")

    parsed = {}
    for k, v in devices.__dict__.items():
        if v.lower() in implementations.keys():
            parsed[k] = implementations[v.lower()]
        else:
            logger.warning(f"Driver implementation for device '{v}' ({k}) not found.")
    return parsed


def list_implementations(modules: List[ModuleType]) -> Dict[str, Any]:
    implementations = {}
    for module in modules:
        impl = _find_implementations(module)
        implementations.update(impl)
    return implementations


def _find_implementations(module: ModuleType) -> Dict[str, Any]:
    impl = {}
    for attrname in dir(module):
        attr = getattr(module, attrname)
        if hasattr(attr, "Model") and callable(attr):
            impl[attrname.lower()] = attr

    return impl



import sys
from . import selector

from . import encoder
from . import motor


impl_modules = [encoder, motor]
implementations = selector.list_implementations(impl_modules)
parsed = selector.parse_device_configuration(impl_modules)
here = sys.modules[__name__]
_ = [setattr(here, k, v) for k, v in parsed.items()]



"""Pure Python tools for NECST."""

import sys
from importlib.metadata import version


# Version Definition
try:
    __version__ = version("neclib")
except:  # noqa: E722
    __version__ = "0.0.0"  # Fallback.


# Environment Variables
class EnvVarName:
    necst_root: str = "NECST_ROOT"
    ros2_ws: str = "ROS2_WS"
    domain_id: str = "ROS_DOMAIN_ID"
    record_root: str = "NECST_RECORD_ROOT"


# Logger configuration
import logging  # noqa: E402

rootLogger = logging.getLogger()
rootLogger.setLevel(logging.DEBUG)
del logging


# Warn Restriction Imposed by Environment
from .interfaces import get_logger  # noqa: E402

logger = get_logger("neclib")
if sys.platform != "linux":
    logger.warning(
        "Device drivers for Interface PCI boards are only supported on Linux."
    )
del logger


# Read Configuration
from .configuration import config, configure  # noqa: F401, E402

config = config


# Perform Time-consuming Downloads
from astropy.time import Time  # noqa: E402

_ = Time.now().ut1  # Will download finals2000A.all (3.4MB) and Leap_Second.dat (1.3KB).
del Time


# Aliases
from .exceptions import *  # noqa: F401, E402, F403


# Submodules
from . import typing  # noqa: F401, E402
from . import units  # noqa: F401, E402


# Subpackages
from . import controllers  # noqa: F401, E402
from . import interfaces  # noqa: F401, E402
from . import parameters  # noqa: F401, E402
from . import recorders  # noqa: F401, E402
from . import safety  # noqa: F401, E402
from . import simulators  # noqa: F401, E402
from . import utils  # noqa: F401, E402
